{-|
  Generic Broker tests. All brokers should satisfy them.
-}

{-# LANGUAGE AllowAmbiguousTypes #-}
{-# LANGUAGE FlexibleContexts #-}
{-# LANGUAGE ScopedTypeVariables #-}

module Test.Integration.Broker
 ( brokerTests
 , initPGMQBroker )
where

import Async.Worker.Broker.PGMQ (PGMQBroker(..))
import Async.Worker.Broker.Types qualified as BT
import Control.Exception (bracket)
import Database.PostgreSQL.Simple qualified as PSQL
import Data.Aeson (ToJSON(..), FromJSON(..), withText)
import Data.Text qualified as T
import Test.Hspec


data TestEnv b =
  TestEnv { broker :: BT.Broker b Message
          , queue  :: BT.Queue }

testQueue :: BT.Queue
testQueue = "test"


data Message =
  Message { text :: String }
  deriving (Show, Eq)
instance ToJSON Message where
  toJSON (Message { text }) = toJSON text
instance FromJSON Message where
  parseJSON = withText "Message" $ \text -> do
    pure $ Message { text = T.unpack text }


initPGMQBroker :: IO (BT.Broker PGMQBroker Message)
initPGMQBroker = do
  let connInfo = PSQL.defaultConnectInfo { PSQL.connectUser = "postgres"
                                         , PSQL.connectDatabase = "postgres" }
  conn <- PSQL.connect connInfo                 
  pure $ BT.initBroker (PGMQBroker conn)

withBroker :: BT.HasBroker b Message => BT.Broker b Message -> (TestEnv b -> IO ()) -> IO ()
withBroker b = bracket (setUpBroker b) tearDownBroker
  where
    -- NOTE I need to pass 'b' again, otherwise GHC can't infer the
    -- type of 'b' (even with 'ScopedTypeVariables' turned on)
    setUpBroker :: BT.HasBroker b Message => BT.Broker b Message -> IO (TestEnv b)
    setUpBroker b' = do
      BT.dropQueue b' testQueue
      BT.createQueue b' testQueue
      
      return $ TestEnv { broker = b'
                       , queue  = testQueue }

    tearDownBroker (TestEnv { broker, queue }) = do
      BT.dropQueue broker queue


-- brokerTests :: BT.Broker broker brokerMessage message msgId => broker -> Spec
-- brokerTests b = sequential $ aroundAll (withBroker b) $ describe "Broker tests" $ do
-- brokerTests :: ( BT.BrokerMessage brokerMessage Message
--                , BT.Broker broker brokerMessage Message msgId )
--               => broker -> Spec
brokerTests :: BT.HasBroker b Message => BT.Broker b Message -> Spec
brokerTests b = sequential $ aroundAll (withBroker b) $ describe "Broker tests" $ do
   it "can send and receive a message" $ \(TestEnv { broker, queue }) -> do
    BT.dropQueue broker queue
    BT.createQueue broker queue
    let msg = Message { text = "test" }
    BT.sendMessage broker queue (BT.toMessage msg)
    msg2 <- BT.readMessageWaiting broker queue
    putStrLn $ "[messageId] " <> show (BT.messageId msg2)
    msg `shouldBe` (BT.toA $ BT.getMessage msg2)

